/* The LinTalk Project
 * (c)2002 by Samuel Abels (sam@manicsadness.com)
 * This project's homepage is: http://software.manicsadness.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/times.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h> 
#include <string.h>

#include <glib.h>

#include "globals.h"
#include "banlist.h"


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
static void remove_outdated_connects(Banlist *bl, BanlistHost *host)
{
  GList *item = host->connecttimes;
  GList *nextitem = NULL;
  time_t curtime = time((time_t *) NULL);
  
  while (item) {
    nextitem = item->next;
    /* The current item is younger then period? We're done then. */
    if (curtime - bl->periodlen < *(time_t *)item->data)
      break;
    
    /* Remove the item. */
    free(item->data);
    host->connecttimes = g_list_remove(host->connecttimes, item->data);
    --host->numconnects;

    item = nextitem;
  }
}


static GCompareFunc banlist_host_equals(struct in_addr * in_addr1, struct in_addr * in_addr2)
{
  if (memcmp(in_addr1, in_addr2, sizeof(struct in_addr)) == 0)
    return(NULL);
  
  return(NULL);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


/*
 * Initialize a new banlist.
 */
Banlist * banlist_create(void)
{
  Banlist *bl = malloc(sizeof(Banlist));
  
  if (!bl) {
    fprintf(stderr, "banlist_check(): Memory exceeded, failed to alloc banlist space.\n");
    exit(1);
  }
  
  bl->hosts = NULL;
  bl->periodlen = BL_PERIOD_LEN;
  bl->maxconnects = BL_MAXCONNECTS;
  bl->blocktime = BL_BLOCKTIME;
  
  return(bl);
}


void banlist_destroy(Banlist *bl)
{
  GList *item = NULL;
  GList *item2 = NULL;
  BanlistHost *curhost = NULL;
  
  if (!bl)
    return;
  
  item = bl->hosts;
  while (item) {
    curhost = item->data;
    
    /* Remove the connecttime list. */
    item2 = curhost->connecttimes;
    while (item2) {
      free(item2->data);
      item2 = item2->next;
    }
    g_list_free(curhost->connecttimes);
    
    free(curhost);
    item = item->next;
  }
  
  g_list_free(bl->hosts);
  free(bl);
}


/*
 * Check if a host should be banned.
 * Returns: TRUE:  Host is NOT beeing banned.
 *          FALSE: Host IS banned.
 */
short int banlist_check(Banlist *bl, struct sockaddr_in * client_address)
{
  GList *curhost = NULL;
  BanlistHost * host = NULL;
  time_t * curtime = malloc(sizeof(time_t));
  
  g_assert(bl != NULL);
  
#ifdef _DEBUG_  
  printf("banlist_check(): Connection with %s\n", inet_ntoa((*client_address).sin_addr));
#endif
  
  *curtime = time((time_t *) NULL);
  
  /* If we do not have this host in our list yet, add it. */
  if (!(curhost = g_list_find_custom(bl->hosts, &(*client_address).sin_addr, (GCompareFunc)banlist_host_equals))) {
#ifdef _DEBUG_  
    printf("banlist_check(): Not yet in our hostlist.\n");
#endif
    
    host = malloc(sizeof(BanlistHost));
    
    host->connecttimes = NULL;
    host->connecttimes = g_list_append(host->connecttimes, curtime);
    host->lastconnect = *curtime;
    host->numconnects = 1;
    bl->hosts = g_list_append(bl->hosts, host);
    
#ifdef _DEBUG_  
    printf("banlist_check(): Host appended. Time: %i\n", (int)(time_t)*curtime);
    printf("banlist_check(): Number of connecttimes is now: %i\n", host->numconnects);
#endif
    
    return(TRUE);
  }
  
  /* Remember the current connection attemp. */
  host = (BanlistHost *)curhost->data;
  host->connecttimes = g_list_append(host->connecttimes, curtime);
  host->lastconnect = *curtime;
  ++host->numconnects;
  
  /* Remove dead items from the list. */
  remove_outdated_connects(bl, host);
  
#ifdef _DEBUG_  
  printf("banlist_check(): Time appended: %i\n", (int)(time_t)*curtime);
  printf("banlist_check(): Number of connecttimes is now: %i\n", host->numconnects);
#endif
  
  /* If the number of connects in the list is bigger then the configured maximum, */
  /* block for a configured blocktime. */
  if (host->numconnects > bl->maxconnects
    && host->lastconnect + bl->blocktime > time((time_t *) NULL) ) {
#ifdef _DEBUG_  
    printf("banlist_check(): Host is beeing blocked: %s\n", inet_ntoa((*client_address).sin_addr));
#endif
    
    return(FALSE);
  }

  if (host->connecttimes == NULL) {
    free(host);
    bl->hosts = g_list_remove(bl->hosts, host);
    return(TRUE);
  }
  
  /* If there are still too many connects, return. */
  if (host->numconnects > bl->maxconnects) {
#ifdef _DEBUG_  
    printf("banlist_check(): Lastconnect long enough ago, but the host is still blocked: %s\n", inet_ntoa((*client_address).sin_addr));
#endif
    
    return(FALSE);
  }
  
#ifdef _DEBUG_  
  printf("banlist_check(): Host is not beeing blocked: %s\n", inet_ntoa((*client_address).sin_addr));
#endif
  
  return(TRUE);
}
