/* The LinTalk Project
 * (c)2002 by Samuel Abels (sam@manicsadness.com)
 * This project's homepage is: http://software.manicsadness.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include "globals.h"
#include "lib_configfile.h"
#include "configfile.h"

extern pthread_mutex_t  mut;
extern Cfg cfg;


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
#define MAXOPTS 500

/*
 * Create the mapping-array to hold the mapping between option names and variables.
 */
static OVPair ** def_optvals(Cfg * cfg)
{
  const OVPair optvals[] = {
    {"internal", "cfgversion", INTEGER, &cfg->version},
    {"local user", "name", STRING, cfg->local.name},
    {"local user", "phone", STRING, cfg->local.phone},
    {"lintalk", "bind to address", STRING, cfg->local.host},
    {NULL, NULL, INTEGER, NULL}
  };
  OVPair ** pairlist = malloc (sizeof(OVPair) * MAXOPTS);
  OVPair *curpair = (OVPair *) optvals;
  int i = 0;

  /* Copy the local array to the global one. */
  while (curpair->option) {
    pairlist[i] = (OVPair *) malloc(sizeof(OVPair));

    pairlist[i]->category = strdup(curpair->category);
    pairlist[i]->option = strdup(curpair->option);
    pairlist[i]->type = curpair->type;
    pairlist[i]->dest = curpair->dest;
    ++i;
    ++curpair;
  }

  pairlist[i] = (OVPair *) malloc(sizeof(OVPair));
  pairlist[i]->option = NULL;
  
  return (pairlist);
}


/*
 * Free the mapping-array.
 */
static void free_optvals(OVPair ** optvals)
{
  OVPair *curpair = optvals[0];
  int i = 0;

  if (!curpair)
    return;

  do {
    free(curpair->category);
    free(curpair->option);
    free(curpair);

    curpair = optvals[++i];
  } while (curpair->option);

  free(curpair);
  free(optvals);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


short int configfile_load(const char *filename, Cfg * cfg)
{
  OVPair ** optvals = NULL;  /* Array to hold the configuration name <-> memory mappings. */
  char config[10000];        /* String to hold the whol configfile. */
  
  optvals = def_optvals(cfg);
  
  /* Read the configfile into one string. */
  if (configfile_read(filename, config, 9999) < 0) {
    fprintf(stderr, "Error: configfile_load: Failed reading the configfile! (%s)\n", filename);
    free_optvals(optvals);
    return(-1);
  }
  
  if (configfile_get_pairs(config, (const OVPair **)optvals) < 0) {
    free_optvals(optvals);
    return(-2);
  }
  
  free_optvals(optvals);
  return(0);
}


/*
 * Save the configuration to a configfile.
 */
short int configfile_save(const char *filename, Cfg * cfg)
{
  OVPair ** optvals = NULL;  /* Array to hold the configuration name <-> memory mappings. */
  char config[10000];        /* String to hold the whol configfile. */
  
  optvals = def_optvals(cfg);
  
  /* Read the old configfile into one string. */
  if (configfile_read(filename, config, 9999) < 0) {
    fprintf(stderr, "Warning: configfile_save(): Failed reading the old configfile! (%s)\n", filename);
    memset(config, 0, sizeof(config));
  }
  
  /* Grab the current configuration into that string. */
  if (configfile_set_pairs(config, (const OVPair **)optvals) < 0) {
    free_optvals(optvals);
    return(-2);
  }
  
  if (configfile_write(filename, config) < 0) {
    fprintf(stderr, "Error: configfile_save(): Failed writing the configfile! (%s)\n", filename);
    free_optvals(optvals);
    return(-1);
  }
  
  free_optvals(optvals);
  return(0);
}


/*
 * Load the user list from a configfile.
 */
int configfile_users_load(void)
{
  User *user = NULL;
  char *usernames[256];
  char *userphone[256];
  char *userhosts[256];
  char *usergroups[256];
  char config[10000] = "";
  char cfgfilename[2048] = "";
  int i = 0;
  
  for (i = 0; i <= 255; i++)
    usernames[i] = NULL;
  
  /* Read config. */
  snprintf (cfgfilename, 2047, "%s/.lintalk/buddies", getenv("HOME"));
  if (configfile_read(cfgfilename, config, 9999) < 0) {
    fprintf(stderr, "Error: configfile_users_load(): Failed to load %s\n", cfgfilename);
    return(-1);
  }

  /* Get other users data. */
  configfile_list_get(config, "user names", usernames);
  configfile_list_get(config, "user phonenumber", userphone);
  configfile_list_get(config, "user hostnames", userhosts);
  configfile_list_get(config, "user groupnames", usergroups);

  /* Put the data to the global userlist. */
  i = 0;
  while (usernames[i] && userphone[i] && userhosts[i] && usergroups[i]) {
    user = user_create(usernames[i], userphone[i], userhosts[i], usergroups[i]);
    free(usernames[i]);
    free(userphone[i]);
    free(userhosts[i]);
    free(usergroups[i]);
    
    pthread_mutex_lock (&mut);
    cfg.users = g_list_append (cfg.users, user);
    pthread_mutex_unlock (&mut);
    
    i++;
  }
  
  return(0);
}


/*
 * Save the user list to a configfile.
 */
int configfile_users_save(void)
{
  GList *item = NULL;
  User *user = NULL;
  
  const char *usernames[256];
  const char *userphone[256];
  const char *userhosts[256];
  const char *usergroups[256];
  char cfgfilename[2048] = "";
  char config[10000] = "";
  int i = 0;
  
  snprintf(cfgfilename, 2047, "%s/.lintalk/buddies", getenv("HOME"));
  memset(config, 0, sizeof(config));
  
  pthread_mutex_lock (&mut);
  
  /* Create an array containing a list of all users. */
  item = cfg.users;
  usernames[i] = NULL;
  userphone[i] = NULL;
  userhosts[i] = NULL;
  usergroups[i] = NULL;
  while (item) {
    user = item->data;
    
    usernames[i] = user->name;
    userphone[i] = user->phone;
    userhosts[i] = user->host;
    usergroups[i] = user->group->name;
    
    i++;
    usernames[i] = NULL;
    userphone[i] = NULL;
    userhosts[i] = NULL;
    usergroups[i] = NULL;
    
    item = item->next;
  }
  
  configfile_list_add(config, "user names", (const char **)usernames);
  configfile_list_add(config, "user phonenumber", (const char **)userphone);
  configfile_list_add(config, "user hostnames", (const char **)userhosts);
  configfile_list_add(config, "user groupnames", (const char **)usergroups);
  
  pthread_mutex_unlock(&mut);
  
  return(configfile_write(cfgfilename, config));
}
