/* The LinTalk Project
 * (c)2002 by Samuel Abels (sam@manicsadness.com)
 * This project's homepage is: http://software.manicsadness.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gtk/gtk.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include "support.h"
#include "globals.h"
#include "gui_preferences.h"
#include "gui_other.h"

extern pthread_mutex_t mut;
extern Cfg cfg;
extern GUI gui;


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
static void on_entry_changed(GtkEntry *entry, gpointer data)
{
  if (*gtk_entry_get_text(entry) == '\0')
    return;
  
  pthread_mutex_lock(&mut);
  strncpy(*(char **)data, gtk_entry_get_text(entry), 2047);
  pthread_mutex_unlock(&mut);
}


/*
 * Change the title label of the current preferences group.
 */
static void preferences_title_set(const char *text)
{
  GtkWidget *label = lookup_widget(gui.preferences, "label_preferences_title");
  char title[2048];
  
  snprintf(title, 2047, "<big><b>%s</b></big>", text);
  /* Change the title label. */
  gtk_label_set_text((GtkLabel *)label, title);
  gtk_label_set_use_markup((GtkLabel *)label, TRUE);
  gtk_widget_set_size_request(label, -1, 35);
  gtk_misc_set_alignment ((GtkMisc *)label, 0.97, 0.5);
  gtk_widget_show(label);
}


/*
 * Show the "About" module.
 */
static GtkWidget * pref_about(const char *name)
{
  GtkWidget *view = NULL;
  char title[2048];
  
  snprintf(title, 2047, "About LinTalk %s", VERSION);
  preferences_title_set(title);
  view = lookup_widget(gui.prefcont, "viewport_about");
  gtk_widget_unparent(view);
  
  return(view);
}


/*
 * Show the "Edit profile" module.
 */
static GtkWidget * pref_editprofile(const char *name)
{
  GtkWidget *view = NULL;
  GtkEntry *entry = NULL;
  
  preferences_title_set(name);
  view = lookup_widget(gui.prefcont, "viewport_editprofile");
  
  /* Fill the forms. */
  pthread_mutex_lock(&mut);
  entry = (GtkEntry *)lookup_widget(gui.prefcont, "entry_prefcont_alias");
  gtk_entry_set_text(entry, cfg.local.name);
  gtk_signal_connect((GtkObject *)entry, "changed", (GtkSignalFunc)on_entry_changed, &cfg.local.name);
  
  entry = (GtkEntry *)lookup_widget(gui.prefcont, "entry_prefcont_phone");
  gtk_entry_set_text(entry, cfg.local.phone);
  gtk_signal_connect((GtkObject *)entry, "changed", (GtkSignalFunc)on_entry_changed, &cfg.local.phone);
  pthread_mutex_unlock(&mut);
  
  gtk_widget_unparent(view);
  return(view);
}


/*
 * Show the "Networking" module.
 */
static GtkWidget * pref_networking(const char *name)
{
  GtkWidget *view = NULL;
  GtkEntry *entry = NULL;
  
  preferences_title_set(name);
  view = lookup_widget(gui.prefcont, "viewport_networking");
  
  /* Fill the forms. */
  pthread_mutex_lock(&mut);
  entry = (GtkEntry *)lookup_widget(gui.prefcont, "entry_prefcont_host");
  gtk_entry_set_text(entry, cfg.local.host);
  gtk_signal_connect((GtkObject *)entry, "changed", (GtkSignalFunc)on_entry_changed, &cfg.local.host);
  pthread_mutex_unlock(&mut);
  
  gtk_widget_unparent(view);
  return(view);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


enum COLUMNS {
  GROUP_COLUMN,
  N_COLUMNS
};

void preferences_switch(GtkTreePath *path)
{
  typedef struct {
    char *name;
    GtkWidget * (* func) (const char *title);
  } Mapping;
  Mapping treemapping[] = {
    {"LinTalk", pref_about},
    {"Networking", pref_networking},
    {"My Profile", pref_editprofile},
    {NULL, NULL}
  };
  GtkWidget *vbox = lookup_widget(gui.preferences, "vbox_preferences_main");
  GtkTreeView *tree = (GtkTreeView *)lookup_widget(gui.preferences, "tree_preferences_group");
  GtkTreeModel *model = gtk_tree_view_get_model((GtkTreeView *)tree);
  GtkTreeIter iter;
  GtkWidget *view = NULL;
  GList *children = NULL;
  char * grpname = NULL;
  int i = 0;
  
  /* Remove the old preferences content. */
  if((children = gtk_container_children((GtkContainer *)vbox))) {
    if (g_list_length(children) == 2) {
      view = g_list_last(children)->data;
      gtk_container_remove((GtkContainer *)vbox, view);
    }
  }
  
  /* No valid tree-row found? */
  if (!path || !gtk_tree_model_get_iter(model, &iter, path)) {
    view = pref_about("LinTalk");
    gtk_box_pack_end((GtkBox *)vbox, view, TRUE, TRUE, 0);
    gtk_widget_show(gui.preferences);
    return;
  }
  
  do {
    gtk_tree_model_get(model, &iter, GROUP_COLUMN, &grpname, -1);
    if (strcmp(treemapping[i].name, grpname) == 0) {
      view = treemapping[i].func(treemapping[i].name);
      g_free(grpname);
    }
  } while(treemapping[++i].name != NULL);
  
  gtk_box_pack_end((GtkBox *)vbox, view, TRUE, TRUE, 0);
  
  gtk_widget_show(gui.preferences);
}


void preferences_init(void)
{
  GtkTreeStore *store;
  GtkWidget *tree = lookup_widget(gui.preferences, "tree_preferences_group");
  GtkTreeIter parentiter;
//  GtkTreeIter childiter;
  GtkTreeViewColumn *column;
  GtkCellRenderer *renderer;

  g_assert(tree != NULL);
  
  /* Create a model.  We are using the store model for now, though we
  * could use any other GtkTreeModel */
  store = gtk_tree_store_new (N_COLUMNS, G_TYPE_STRING);
  gtk_tree_view_set_model(GTK_TREE_VIEW (tree), GTK_TREE_MODEL (store));
  
  /* Create a cell render and arbitrarily make it red for demonstration purposes */
  renderer = gtk_cell_renderer_text_new ();
  g_object_set(G_OBJECT(renderer), "foreground", "black", NULL);
  
  /* Create a column, associating the "text" attribute of the
  * cell_renderer to the first column of the model */
  column = gtk_tree_view_column_new_with_attributes("Users", renderer, "text", GROUP_COLUMN, NULL);
  
  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW (tree), column);
  
  /* Group "LinTalk". */
  gtk_tree_store_append(store, &parentiter, NULL);
  gtk_tree_store_set(store, &parentiter, GROUP_COLUMN, "LinTalk", -1);
  
  /* Subgroup "Networking". */
//  gtk_tree_store_append(store, &childiter, &parentiter);
//  gtk_tree_store_set(store, &childiter, GROUP_COLUMN, "Networking", -1);
  
  /* Group "Edit Profile". */
  gtk_tree_store_append(store, &parentiter, NULL);
  gtk_tree_store_set(store, &parentiter, GROUP_COLUMN, "My Profile", -1);
  
  /* Group "Conversations". */
//  gtk_tree_store_append(store, &parentiter, NULL);
//  gtk_tree_store_set(store, &parentiter, GROUP_COLUMN, "Conversations", -1);

  gtk_tree_view_expand_all((GtkTreeView *)tree);
}

#undef COLUMNS
