/* The LinTalk Project
 * (c)2002 by Samuel Abels (sam@manicsadness.com)
 * This project's homepage is: http://software.manicsadness.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */  
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <gtk/gtk.h>
#include "globals.h"
#include "gui_talk.h"
#include "users.h"
#include "support.h"
#include "gui_other.h"
#include "callbacks.h"
extern pthread_mutex_t mut;
extern Cfg cfg;

/***************************************************************************************
 * BELOW FOLLOW THE STATICS
 ***************************************************************************************/
enum COLORS {
  BLUE,
  RED
};

static void history_insert(GtkTextView * history, const char *username, const char *text, int color)
{
  GtkTextBuffer *buffer = NULL;
  GtkTextTagTable *tagtable = NULL;
  GtkTextTag *tag = NULL;
  GtkTextIter end;
  
  buffer = gtk_text_view_get_buffer(history);
  g_return_if_fail(buffer != NULL);
  
  /* Check if there are already color-tags in the buffer. */
  tagtable = gtk_text_buffer_get_tag_table(buffer);
  
  /* Create a text-tag for the correct color. */
  switch (color) {
  case BLUE:
    if (!(tag = gtk_text_tag_table_lookup(tagtable, "blue_foreground")))
      tag = gtk_text_buffer_create_tag(buffer, "blue_foreground", "foreground", "blue", NULL);
    break;
  case RED:
    if (!(tag = gtk_text_tag_table_lookup(tagtable, "red_foreground")))
      tag = gtk_text_buffer_create_tag(buffer, "red_foreground", "foreground", "red", NULL);
    break;
  }
  
  /* Append the text to the buffer. */
  gtk_text_buffer_get_end_iter(buffer, &end);
  gtk_text_buffer_insert_with_tags(buffer, &end, username, -1, tag, NULL);
  gtk_text_buffer_insert_with_tags(buffer, &end, ": ", -1, tag, NULL);
  gtk_text_buffer_insert_with_tags(buffer, &end, text, -1, tag, NULL);
  gtk_text_buffer_insert_with_tags(buffer, &end, "\n", -1, tag, NULL);
}


/*
 * Return the content of a textview widget.
 */
static gchar * get_textview_text(GtkTextView *tv)
{
  GtkTextBuffer *buffer = NULL;
  GtkTextIter start;
  GtkTextIter end;
  
  g_return_val_if_fail(tv != NULL, NULL);
  
  buffer = gtk_text_view_get_buffer(tv);
  
  g_return_val_if_fail(buffer != NULL, NULL);
  
  gtk_text_buffer_get_start_iter(buffer, &start);
  gtk_text_buffer_get_end_iter(buffer, &end);
  
  return(gtk_text_buffer_get_text((GtkTextBuffer *)buffer, &start, &end, FALSE));
}


/*
 * Delete the content of a textview widget.
 */
static void del_textview_text(GtkTextView *tv)
{
  GtkTextBuffer *buffer = NULL;
  GtkTextIter start;
  GtkTextIter end;
  
  g_return_if_fail(tv != NULL);
  
  buffer = gtk_text_view_get_buffer(tv);
  
  g_return_if_fail(buffer != NULL);
  
  gtk_text_buffer_get_start_iter(buffer, &start);
  gtk_text_buffer_get_end_iter(buffer, &end);
  gtk_text_buffer_delete((GtkTextBuffer *)buffer, &start, &end);
}
/***************************************************************************************
 * END OF STATICS
 ***************************************************************************************/


/*
 * Create the "Talk" widget.
 */
GtkWidget *talk_create(User * user)
{
  GtkWidget *talk = create_talk();
  GtkWidget *textview_talk_history = NULL;
  GtkWidget *textview_talk_text = NULL;
  GtkWidget *button_talk_send = NULL;
  GtkWidget *button_talk_cancel = NULL;
  char title[2048];
  
  if (!talk)
    return(NULL);

  /* Find widgets. */
  textview_talk_history = (GtkWidget *)lookup_widget(talk, "textview_talk_history");
  textview_talk_text = (GtkWidget *)lookup_widget(talk, "textview_talk_text");
  button_talk_send = (GtkWidget *)lookup_widget(talk, "button_talk_send");
  button_talk_cancel = (GtkWidget *)lookup_widget(talk, "button_talk_cancel");
  
  /* Set the window title. */
  snprintf(title, 2047, _("Talk to %s"), user->name);
  gtk_window_set_title(GTK_WINDOW(talk), title);
  
  /* Connect signals. */
  gtk_signal_connect(GTK_OBJECT(textview_talk_text), "key_press_event", GTK_SIGNAL_FUNC(on_textview_talk_text_key_pressed), user);
  gtk_signal_connect(GTK_OBJECT(button_talk_cancel), "clicked", GTK_SIGNAL_FUNC(on_button_talk_cancel_clicked), user);
  gtk_signal_connect(GTK_OBJECT(button_talk_send), "clicked", GTK_SIGNAL_FUNC(on_button_talk_send_clicked), user);
  gtk_signal_connect(GTK_OBJECT(talk), "destroy", GTK_SIGNAL_FUNC(on_talk_destroy), user);

  return(talk);
}


/*
 * Will be executed after the user hit the "Send" button.
 */
void talk_send(User * user)
{
  const gchar *text = NULL;
  GtkWidget *history = NULL;
  GtkWidget *outtext = NULL;

  g_return_if_fail(user != NULL);
  g_return_if_fail(user->talk != NULL);

  history = lookup_widget(user->talk, "textview_talk_history");
  outtext = lookup_widget(user->talk, "textview_talk_text");
  text = get_textview_text((GtkTextView *)outtext);

  if (*text == '\0')
    return;

  pthread_mutex_lock(&mut);
  
  /* Make sure the user has not yet been deleted. */
  if (!user) {
    fprintf(stderr, _("talk_send(): User exists no more!\n"));
    return;
  }
  
  /* Add the message to the out-queue. */
  user->outqueue = g_list_append(user->outqueue, strdup(text));

  /* Update the history. */
  history_insert((GtkTextView *)history, cfg.local.name, text, BLUE);
  pthread_mutex_unlock(&mut);
  
  del_textview_text((GtkTextView *)outtext);
}


/*
 * Will be executed after the user doubleclicked a buddy.
 */
void talk_to(User * user)
{
  if (!user)
    return;
  
  /* Does this widget already exist? */
  if (user->talk)
    return;
  
  /* Show a new "Talk" widget. */
  user->talk = talk_create(user);
  talk_update(user);
  gtk_widget_show(user->talk);
  
  return;
}


/*
 * Update the whole history.
 */
void talk_update(User * user)
{
  Historyentry *histentry = NULL;
  GtkWidget *history = NULL;
  GList *item = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter start, end;

  g_return_if_fail(user != NULL);
  g_return_if_fail(user->talk != NULL);

  history = lookup_widget(user->talk, "textview_talk_history");

  /* Empty out the history. */
  buffer = gtk_text_view_get_buffer((GtkTextView *)history);
  gtk_text_buffer_get_start_iter(buffer, &start);
  gtk_text_buffer_get_end_iter(buffer, &end);
  gtk_text_buffer_delete(buffer, &start, &end);

  /* Insert all items. */
  item = user->history;
  while (item != NULL) {
    histentry = item->data;
    
    if ( histentry->user == 0 )
      history_insert((GtkTextView *)history, histentry->username, histentry->text, BLUE);
    else
      history_insert((GtkTextView *)history, histentry->username, histentry->text, RED);
    
    item = item->next;
  }
}


/*
 * Insert the last item of the user's history into the gui.
 */
void talk_update_last(User * user)
{
  Historyentry *histentry = NULL;
  GtkWidget *history = NULL;

  g_return_if_fail(user != NULL);
  g_return_if_fail(user->history != NULL);

  gdk_threads_enter();

  /* If the window is not open, open it. */
  if (!user->talk) {
    user->talk = talk_create(user);
    talk_update(user);
    gtk_widget_show(user->talk);
    gdk_threads_leave();
    return;
  }

  /* Insert last item of the history. */
  histentry = (Historyentry *)g_list_last(user->history)->data;
  if ( histentry->user == 0 )
    history_insert((GtkTextView *)history, histentry->username, histentry->text, BLUE);
  else
    history_insert((GtkTextView *)history, histentry->username, histentry->text, RED);

  gdk_threads_leave();
}
