/* The LinTalk Project
 * (c)2002 by Samuel Abels (sam@manicsadness.com)
 * This project's homepage is: http://software.manicsadness.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gtk/gtk.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include "globals.h"
#include "groups.h"
#include "users.h"
#include "gui_users.h"
#include "support.h"
#include "gui_userlist.h"

extern pthread_mutex_t mut;
extern Cfg cfg;
extern GUI gui;


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
typedef struct {
  void *userdata;
  int column;
  GtkTreeIter *iter;
  gboolean found;
} ForEachCmp;

static gboolean pointer_same(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, ForEachCmp *rowdata)
{
  void * data = NULL;
  
  gtk_tree_model_get(model, iter, rowdata->column, &data, -1);
  
  if (rowdata->userdata == data) {
    memcpy(rowdata->iter, iter, sizeof(GtkTreeIter));
    rowdata->found = TRUE;
    return(TRUE);
  }
  
  return(FALSE);
}


/*
 * Find the tree row/column that contains "userdata"
 * and store the position in "iter".
 * Returns TRUE on success, otherwise FALSE.
 */
static gboolean find_treeview_row(GtkTreeModel *model, GtkTreeIter * iter, int column, void * userdata)
{
  ForEachCmp data;
  
  data.userdata = userdata;
  data.column = column;
  data.iter = iter;
  data.found = FALSE;
  
  gtk_tree_model_foreach(model, (GtkTreeModelForeachFunc)pointer_same, &data);
  
  return(data.found);
}


static gboolean on_userlist_drag_drop(GtkTreeView *tree,
                                      GdkDragContext *context,
                                      gint x,
                                      gint y,
                                      guint time,
                                      gpointer user_data)
{
  GtkTreeSelection *selection = NULL;
  GtkTreePath *parentpath = NULL;
  GtkTreePath *siblingpath = NULL;
  GtkTreePath *draggedpath = NULL;
  GtkTreeModel *model = NULL;
  GtkTreeIter iter;
  Group * destgroup = NULL;
  Group * siblinggroup = NULL;
  Group * draggedgroup = NULL;
  User * siblinguser;
  User * draggeduser;
  int type = 0;
  
  /* Get the group the user was dragged to. */
  gtk_tree_view_get_path_at_pos(tree, x, y, &parentpath, NULL, NULL, NULL);
  if (!parentpath)
    return(TRUE);
  if (gtk_tree_path_get_depth(parentpath) < 1)
    return(TRUE);
  if (!gtk_tree_path_up(parentpath))
    parentpath = NULL;
  if (gtk_tree_path_get_depth(parentpath) < 1)
    parentpath = NULL;
  
  /* Get the sibling. */
  gtk_tree_view_get_path_at_pos(tree, x, y, &siblingpath, NULL, NULL, NULL);
  
  /* Get the dragged item. */
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
  if (!selection)
    return(TRUE);
  draggedpath = (GtkTreePath *)gtk_tree_selection_get_selected_rows(selection, &model)->data;
  if (!draggedpath)
    return(TRUE);
  
  /* Now we have all the paths that we need. */
  model = (GtkTreeModel *)gtk_tree_view_get_model((GtkTreeView *)tree);
  if (!model)
    return(TRUE);
  
  /* Get the destination group's struct. */
  if (parentpath) {
    if (!gtk_tree_model_get_iter(model, &iter, parentpath))
      return(TRUE);
    gtk_tree_model_get(model, &iter, TYPE_COLUMN, &type, DATA_COLUMN, &destgroup, -1);
    if (type != GROUP_ROW)
      return(TRUE);
  }
  if (!parentpath && !(destgroup = group_find_from_number(1)))
    return(TRUE);

  /* Get the sibling user/group's struct. */
  if (!gtk_tree_model_get_iter(model, &iter, siblingpath))
    return(TRUE);
  gtk_tree_model_get(model, &iter, TYPE_COLUMN, &type, DATA_COLUMN, &siblinguser, -1);
  if (type != USER_ROW) {
    siblinguser = NULL;
    gtk_tree_model_get(model, &iter, DATA_COLUMN, &siblinggroup, -1);
  }
  
  /* Get the dragged user/group's struct. */
  if (!gtk_tree_model_get_iter(model, &iter, draggedpath))
    return(TRUE);
  gtk_tree_model_get(model, &iter, TYPE_COLUMN, &type, DATA_COLUMN, &draggeduser, -1);
  if (type != USER_ROW) {
    draggeduser = NULL;
    gtk_tree_model_get(model, &iter, DATA_COLUMN, &draggedgroup, -1);
  }
/*  
  printf("Parent: %s, Sibling: %s, Dragged: %s\n", destgroup->name,
          siblinguser ? siblinguser->name : siblinggroup->name,
          draggeduser ? draggeduser->name : draggedgroup->name);
  */
  return(FALSE);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


void userlist_create(void)
{
  GtkTreeStore *store;
  GtkWidget *tree = lookup_widget(gui.lintalk, "tree_users");
  GtkTreeViewColumn *column;
  GtkCellRenderer *renderer;

  g_assert(tree != NULL);
  
  /* Create a model and connect signals. */
  store = gtk_tree_store_new(N_COLUMNS, G_TYPE_STRING, G_TYPE_POINTER, G_TYPE_INT);
  gtk_tree_view_set_model(GTK_TREE_VIEW(tree), GTK_TREE_MODEL(store));
  gtk_signal_connect(GTK_OBJECT(tree), "drag-drop", (GtkSignalFunc)on_userlist_drag_drop, NULL);
  
  /* Create a cell render and arbitrarily make it red for demonstration purposes */
  renderer = gtk_cell_renderer_text_new();
  g_object_set(G_OBJECT(renderer), "foreground", "black", NULL);
  
  /* Create a column, associating the "text" attribute of the
  * cell_renderer to the first column of the model */
  column = gtk_tree_view_column_new_with_attributes("Users", renderer, "text", USERS_COLUMN, NULL);
  
  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
}


/*
 * Update the userlist in the GUI.
 */
void userlist_update(void)
{
  GtkWidget *tree = lookup_widget(gui.lintalk, "tree_users");
  GtkTreeIter parentiter;  /* Parent iter */
  GtkTreeIter childiter;   /* Child iter  */
  GtkTreeStore *store = NULL;
  GList *item = NULL;
  Group *group = NULL;
  User *user = NULL;
  gboolean inlist = FALSE;
  
  if(!tree)
    return;
  
  store = (GtkTreeStore *)gtk_tree_view_get_model((GtkTreeView *)tree);
  
  if(!store)
    return;
  
  /* Lock everything. */
  pthread_mutex_lock (&mut);

  /* Walk through the userlist. */
  item = cfg.users;
  while (item) {
    user = item->data;
    
    /* Find the user row. */
    inlist = find_treeview_row((GtkTreeModel *)store, &childiter, DATA_COLUMN, user);
    /* If the user is already in the list and he is not available, remove him. */
    if (inlist && user->status == OFFLINE) {
      gtk_tree_store_remove(store, &childiter);
      user->updategui = FALSE;
      item = item->next;
      continue;
    }
    
    /* If the user's category has changed (updategui == 1), remove him first. */
    if (inlist && user->updategui) {
      gtk_tree_store_remove(store, &childiter);
      user->updategui = FALSE;
      inlist = FALSE;
    }
    
    /* No group? Then do nothing else. */
    if (user->group == NULL) {
      item = item->next;
      continue;
    }
    
    /* Group = root? Then add the user without any category. */
    if (!inlist && user->group->number == 0 && user->status != OFFLINE) {
      gtk_tree_store_append(store, &childiter, NULL);
      gtk_tree_store_set(store, &childiter, USERS_COLUMN, user->name, DATA_COLUMN, user, TYPE_COLUMN, USER_ROW, -1);
      item = item->next;
      continue;
    }
    
    if (user->group->number == 0) {
      item = item->next;
      continue;
    }
    
    /* Check if the user's group already exists int the gui and get the parent iter. */
    /* Otherwise, create it. */
    group = user->group;
    if (!find_treeview_row((GtkTreeModel *)store, &parentiter, DATA_COLUMN, group)) {
      gtk_tree_store_append(store, &parentiter, NULL);
      gtk_tree_store_set(store, &parentiter, USERS_COLUMN, group->name, DATA_COLUMN, group, TYPE_COLUMN, GROUP_ROW, -1);
    }
    
    /* If the user is already in the list and he is still available, update his name. */
    if (inlist && user->status != OFFLINE)
      gtk_tree_store_set(store, &childiter, USERS_COLUMN, user->name, DATA_COLUMN, user, TYPE_COLUMN, USER_ROW, -1);
    
    /* If the user is not in the list but he is available, add him. */
    if (!inlist && user->status != OFFLINE) {
      GtkTreePath *path = NULL;
      
      gtk_tree_store_append(store, &childiter, &parentiter);
      gtk_tree_store_set(store, &childiter, USERS_COLUMN, user->name, DATA_COLUMN, user, TYPE_COLUMN, USER_ROW, -1);
      path = gtk_tree_model_get_path((GtkTreeModel *)store, &childiter);
      gtk_tree_view_expand_to_path((GtkTreeView *)tree, path);
      gtk_tree_path_free(path);
    }

    item = item->next;
  }
  
  /* Unlock stuff. */
  pthread_mutex_unlock (&mut);
}


/*
 * Delete all selected users.
 */
void userlist_delete_selected(void)
{
  GtkTreeView *tree = (GtkTreeView *)lookup_widget(gui.lintalk, "tree_users");
  GtkTreeSelection *selection = NULL;
  GtkTreeModel *model = NULL;
  GtkTreePath *path = NULL;
  GtkTreeIter iter;
  GList *item = NULL;
  User *user = NULL;
  Group *group = NULL;
  int type = GROUP_ROW;
  
  g_assert(tree != NULL);

  /* Remove all currently selected users. */
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
  
  if (!selection)
    return;

  item = gtk_tree_selection_get_selected_rows(selection, &model);
  while (item) {
    /* Find row. */
    path = (GtkTreePath *)item->data;
    if (!gtk_tree_model_get_iter(model, &iter, path)) {
      item = item->next;
      continue;
    }
    
    /* Don't remove non-empty categories. */
    if (gtk_tree_model_iter_has_child(model, &iter)) {
      item = item->next;
      continue;
    }

    /* Find the row type. */
    gtk_tree_model_get((GtkTreeModel *)model, &iter, TYPE_COLUMN, &type, -1);
    
    /* Is it an empty category? Remove it from the GUI. */
    switch (type) {
    case GROUP_ROW:
      if (gtk_tree_path_get_depth(path) == 1) {
        gtk_tree_store_remove((GtkTreeStore *)model, &iter);
      }
      break;
      
    case USER_ROW:
      if (find_treeview_row((GtkTreeModel *)model, &iter, DATA_COLUMN, group)) {
        gtk_tree_store_remove((GtkTreeStore *)model, &iter);
        user->group = NULL;
      }
      break;
    }
    
    item = item->next;
  }
}

#undef COLUMNS
#undef ROW_TYPES
