/* The LinTalk Project
 * (c)2002 by Samuel Abels (sam@manicsadness.com)
 * This project's homepage is: http://software.manicsadness.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include "globals.h"
#include "support.h"
#include "users.h"
#include "gui_users.h"
#include "gui_userlist.h"
#include "gui_other.h"
#include "gui_popup.h"
#include "network.h"

extern pthread_mutex_t mut;
extern Cfg cfg;
extern GUI gui;

enum USER_COLUMNS {
  USER_COLUMN_FIRST,
  USER_COLUMN_DATA,
  USER_COLUMN_NUM
};


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
typedef struct {
  void *userdata;
  int column;
  GtkTreeIter *iter;
  gboolean found;
} ForEachCmp;

static gboolean pointer_same(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, ForEachCmp *rowdata)
{
  void * data = NULL;
  
  gtk_tree_model_get(model, iter, rowdata->column, &data, -1);
  
  if (rowdata->userdata == data) {
    memcpy(rowdata->iter, iter, sizeof(GtkTreeIter));
    rowdata->found = TRUE;
    return(TRUE);
  }
  
  return(FALSE);
}


/*
 * Find the tree row/column that contains "userdata"
 * and store the position in "iter".
 * Returns TRUE on success, otherwise FALSE.
 */
static gboolean find_treeview_row(GtkTreeModel *model, GtkTreeIter * iter, int column, void * userdata)
{
  ForEachCmp data;
  
  data.userdata = userdata;
  data.column = column;
  data.iter = iter;
  data.found = FALSE;
  
  gtk_tree_model_foreach(model, (GtkTreeModelForeachFunc)pointer_same, &data);
  
  return(data.found);
}


static void update_availablebuddies(void)
{
  GtkWidget *tree = lookup_widget(gui.newuser, "tree_newuser_knownbuddies");
  GtkTreeSelection *selection = NULL;
  GtkTreeStore *store = NULL;
  GtkTreeIter iter;
  GList *item = NULL;
  User *user = NULL;
  int inlist = FALSE;
  
  if(!tree)
    return;
  
  store = (GtkTreeStore *)gtk_tree_view_get_model((GtkTreeView *)tree);
  selection = (GtkTreeSelection *)gtk_tree_view_get_selection((GtkTreeView *)tree);
  
  if (!store || !selection)
    return;
  
  /* Lock everything. */
  pthread_mutex_lock (&mut);

  /* Walk through the userlist. */
  item = cfg.users;
  while (item) {
    user = item->data;
    
    /* Get the user's iter. */
    inlist = find_treeview_row((GtkTreeModel *)store, &iter, USER_COLUMN_DATA, user);
    
    /* If the user is already in the list, just update his name. */
    if (inlist) {
      gtk_tree_store_set(store, &iter, USER_COLUMN_FIRST, user->name, -1);
      item = item->next;
      continue;
    }
    
    /* Add a new user. */
    gtk_tree_store_append(store, &iter, NULL);
    gtk_tree_store_set(store, &iter, USER_COLUMN_FIRST, user->name, USER_COLUMN_DATA, user, -1);
    gtk_tree_selection_select_iter(selection, &iter);
    
    item = item->next;
  }
  
  /* Unlock stuff. */
  pthread_mutex_unlock (&mut);
}


static void update_availablegroups(void)
{
  GtkWidget *tree = lookup_widget(gui.newuser, "tree_newuser_groups");
  GtkTreeSelection *selection = NULL;
  GtkTreeStore *store = NULL;
  GtkTreeIter iter;
  Group *group = NULL;
  int inlist = FALSE;
  int i = 0;
  
  if(!tree)
    return;
  
  store = (GtkTreeStore *)gtk_tree_view_get_model((GtkTreeView *)tree);
  selection = (GtkTreeSelection *)gtk_tree_view_get_selection((GtkTreeView *)tree);
  
  if (!store || !selection)
    return;
  
  /* Lock everything. */
  pthread_mutex_lock(&mut);

  /* Walk through the grouplist. */
  i = -1;
  while ((group = group_find_from_number(++i))) {
    /* Get the child iter. */
    inlist = find_treeview_row((GtkTreeModel *)store, &iter, USER_COLUMN_DATA, group);
    
    /* If the group is already in the list do nothing. */
    if (inlist)
      continue;
    
    gtk_tree_store_append(store, &iter, NULL);      
    gtk_tree_store_set(store, &iter, USER_COLUMN_FIRST, group->name, USER_COLUMN_DATA, group, -1);
    gtk_tree_selection_select_iter(selection, &iter);
  }
  
  /* Unlock stuff. */
  pthread_mutex_unlock(&mut);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


/*
 * Create the list of available users and groups.
 */
void newuser_lists_create(void)
{
  GtkTreeStore *store;
  GtkWidget *tree = lookup_widget(gui.newuser, "tree_newuser_knownbuddies");
  GtkTreeSelection * selection = NULL;
  GtkTreeViewColumn *column;
  GtkCellRenderer *renderer;

  g_assert(tree != NULL);
  
  /* ----------------------------- */
  /* Create the user-list first.   */
  /* ----------------------------- */
  
  /* Create a model. */
  store = gtk_tree_store_new (USER_COLUMN_NUM, G_TYPE_STRING, G_TYPE_POINTER);
  gtk_tree_view_set_model(GTK_TREE_VIEW(tree), GTK_TREE_MODEL(store));
  
  /* Create the "Users" column. */
  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes("Users", renderer, "text", USER_COLUMN_FIRST, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
  
  /* Set the selection type to "multiple". */
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
  gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);
  
  /* ----------------------------- */
  /* Now create the category-list. */
  /* ----------------------------- */
  tree = lookup_widget(gui.newuser, "tree_newuser_groups");

  g_assert(tree != NULL);
  
  /* Create a model. */
  store = gtk_tree_store_new(USER_COLUMN_NUM, G_TYPE_STRING, G_TYPE_POINTER);
  gtk_tree_view_set_model(GTK_TREE_VIEW(tree), GTK_TREE_MODEL(store));
  
  /* Create the "Groups" column. */
  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes("Groups", renderer, "text", USER_COLUMN_FIRST, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
}


/*
 * Update the list of available users and groups.
 */
void newuser_lists_update(void)
{
  update_availablebuddies();
  update_availablegroups();
}


/*
 * Will be called when the user confirmed the "Add Buddy" dialog.
 */
void newuser_add_selected(void)
{
  GtkTreeView *tree = (GtkTreeView *)lookup_widget(gui.newuser, "tree_newuser_groups");
  GtkTreeModel *model = NULL;
  GtkTreeSelection *selection = NULL;
  GtkTreeIter iter;
  GList *item = NULL;
  Group *group = NULL;
  User *user = NULL;
  
  g_assert(tree != NULL);

  /* Get the seleted group's id. */
  if (!(selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree))))
    return;
  
  if (!(item = gtk_tree_selection_get_selected_rows(selection, &model))) {
    popup("No group selected.\n\nPlease select a group to add users to your buddylist.\n", GTK_MESSAGE_ERROR);
    return;
  }
  
  /* Find the first selected group. */
  if (!gtk_tree_model_get_iter(model, &iter, (GtkTreePath *)item->data)) {
    popup("Internal fatal error.\nFailed to get group iterator.\n\nIf this is the newest release of this program,\nplease report a bug to the author.\n", GTK_MESSAGE_ERROR);
    return;
  }
  gtk_tree_model_get(model, &iter, USER_COLUMN_DATA, &group, -1);
  
  /* Walk through all selected users and set their group id accordingly. */
  tree = (GtkTreeView *)lookup_widget(gui.newuser, "tree_newuser_knownbuddies");
  g_return_if_fail(tree != NULL);
  
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
  g_return_if_fail(selection != NULL);
  
  pthread_mutex_lock(&mut);
  item = gtk_tree_selection_get_selected_rows(selection, &model);
  while (item) {
    /* Find the user. */
    if (!gtk_tree_model_get_iter(model, &iter, (GtkTreePath *)item->data)) {
      item = item->next;
      continue;
    }
    gtk_tree_model_get(model, &iter, USER_COLUMN_DATA, &user, -1);
    
    user->group = group;
    user->updategui = TRUE;
    
    item = item->next;
  }
  pthread_mutex_unlock(&mut);
  
  /* Force list updates. */
  userlist_update();
  update_availablebuddies();
}


/*
 * Will be executed to add a user to the "Available Buddies" list.
 */
void newuser_add(GtkWidget *adduser)
{
  GtkEntry * entry = (GtkEntry *)lookup_widget(adduser, "entry_newuser_host");
  GList *item = NULL;
  GList *hosts = NULL;
  char *host = NULL;
  User *newuser = NULL;
  User *curuser = NULL;
  
  /* The user must enter a hostname. */
  if (*gtk_entry_get_text(entry) == '\0')
    return;
  
  /* If the user's IP is your own, just quit. */
  if (strcmp(gtk_entry_get_text(entry), cfg.local.host) == 0) {
    popup("You have entered your own IP.\n\nIt's not possible to connect to yourself.\n", GTK_MESSAGE_ERROR);
    return;
  }
  
  /* Get data from the gui into a new user. */
  newuser = user_create("", "", gtk_entry_get_text(entry), NULL);

  /* The user must be available. */
  receive_who(newuser, newuser->host);
  if (newuser->status == OFFLINE) {
    user_free(newuser);
    popup("Sorry, can't contact host!\n\nPlease check if LinTalk is online on the remote host.\n", GTK_MESSAGE_ERROR);
    return;
  }
  
  /* If the user is already in the list, just quit. */
  pthread_mutex_lock(&mut);
  if (user_find_from_host(cfg.users, newuser->host)) {
    pthread_mutex_unlock(&mut);
    user_free(newuser);
    popup("The user was already in our list.\n\nTo add him to your buddylist select him in the userlist and click the \"Add Selected\" button.\n", GTK_MESSAGE_WARNING);
    /* Cleanup hostlist copy. */
    g_list_free(hosts);
    return;
  }
  
  /* Create a list of all known hosts (must copy to be threadsafe). */
  item = cfg.users;
  while (item != NULL) {
    curuser = item->data;
    hosts = g_list_append(hosts, strdup(curuser->host));
    item = item->next;
  }
  pthread_mutex_unlock(&mut);
  
  /* Tell this new host to every knownhost... */
  item = hosts;
  while (item) {
    host = item->data;
    
    /* Don't connect to yourself. */
    if (strcmp(host, cfg.local.host) == 0) {
      item = item->next;
      continue;
    }
    
    /* Notify neighbor. */
    send_notify(newuser, host);
    /* Cleanup. */
    free(host);
    item = item->next;
  }
  g_list_free(hosts);
  
  /* ...and put it in our own userlist. */
  pthread_mutex_lock(&mut);
  cfg.users = g_list_append(cfg.users, newuser);
  pthread_mutex_unlock(&mut);
  
  gtk_entry_set_text((GtkEntry*)lookup_widget(adduser, "entry_newuser_host"), "");
  
  update_availablebuddies();
}


/*
 * Will be executed to add a group to the "Available Groups" list.
 */
void newuser_add_group(GtkWidget *adduser)
{
  GtkEntry * entry = (GtkEntry*)lookup_widget(adduser, "entry_newuser_group");
  
  if (*gtk_entry_get_text(entry) == '\0')
    return;
  
  if (group_find_from_name(gtk_entry_get_text(entry))) {
    popup("This group does already exist.\n", GTK_MESSAGE_ERROR);
    return;
  }
  
  /* Get data from the gui into a new group. */
  group_add(strdup(gtk_entry_get_text(entry)));
  
  gtk_entry_set_text(entry, "");
  
  update_availablegroups();
}

#undef COLUMNS
