/* lib_configfile
 * (c)2002, 2003 by Samuel Abels (sam@manicsadness.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* #define _DEBUG_ */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "lib_configfile.h"

/***************************************************************************************
 * STATICS
 ***************************************************************************************/
/*
 * Return the last item of an doubly linked list
 */
static DLL * dll_last (DLL * list)
{
  if (list == NULL)
    return (NULL);
  
  while (list->next != NULL)
    list = list->next;
  
  return (list);
}

/*
 * Append an item to the doubly linked list
 */
static DLL * dll_append (DLL * list, void * data)
{
  DLL *item = malloc(sizeof(DLL));
  DLL *lastitem = dll_last(list);
  
  item->data = data;
  item->next = NULL;
  
  if (lastitem == NULL) {
    item->prev = NULL;
    return (item);
  } else {
    item->prev = lastitem;
    lastitem->next = item;
  }
  
  return (list);
}


static int dll_position (DLL * list, DLL * needleitem)
{
  int i = 0;
  DLL *item = list;
  
  while ( item ) {
    if ( item == needleitem )
      return (i);
    ++i;
    item = item->next;
  }
  
  return (-1);
}


static DLL * dll_insert (DLL * list, void * data, int position)
{
  DLL *curitem = list;
  DLL *newitem = NULL;
  DLL *nextitem = NULL;
  int i = 0;
  
  /* Find the right position */
  for (i = 0; curitem && i != position; i++)
    curitem = curitem->next;
  
  nextitem = curitem->next;
  
  /* Create the new item */
  newitem = malloc (sizeof (DLL));
  newitem->data = data;
  newitem->prev = curitem;
  newitem->next = nextitem;
  
  /* This could have been the first entry. */
  if ( !curitem )
    return (newitem);
  
  /* Link the current and the next item to the new one */
  curitem->next = newitem;
  if ( nextitem )
    nextitem->prev = newitem;
  
  return (list);
}


/*
 * Free a doubly linked list.
 */
static DLL * dll_free (DLL * list)
{
  DLL *item = list;
  DLL *current = NULL;

  while (item) {
    current = item;
    item = item->next;

    free (current);
  }

  return (NULL);
}


/*
 * This function will kill spaces at the start/end/both of source. Spaces are:
 * single space, tab, vertical tab, form feed, carriage return, or newline
 * loc: 0-start 1-end 2-both
 */
static void wipe_spaces (char * source, int loc)
{
  char *psource = NULL;
  char *destination = NULL;
  char *pdestination = NULL;
  
  /* Must have a source */
  if ( source == NULL )
    return;
  
  /* Kill spaces at the beginning of a string */
  if ( loc == 0 || loc == 2 ) {
    /* Create a copy of the source string */
    pdestination = destination = strdup (source);
    
    /* Skip spaces */
    while ( *pdestination == ' ' )
      pdestination++;
    
    /* Copy the string back */
    strcpy (source, pdestination);
    
    free (destination);
  }
  
  /* Kill spaces at the end of a string. */
  if ( loc == 1 || loc == 2 ) {
    for (psource = source + strlen (source) - 1; *psource != '\0'; psource--) {
      if ( isspace (*psource) )
        *psource='\0';
      else
        break;
    }
  }
  
  return;
}


/*
 * Like strdup, but ends on "\n" too.
 */
static char * get_line_from_var (const char * var)
{
  char *line = NULL;
  char *pline = NULL;
  
  pline = line = malloc (strlen (var) + 1);
  
  while ( *var != '\n' && *var != '\0' )
    *(pline++) = *(var++);
  
  *pline = '\0';
  
  return (line);
}


/*
 * Get the complete config to a glist (so the rows are separated).
 */
static DLL * get_config_to_dll (const char * config)
{
  DLL *dll = NULL;
  char *pconfig = (char *)config;
  
  /* Parse through the config variable */
  while ( *pconfig != '\0' ) {
    dll = dll_append (dll, get_line_from_var (pconfig));
    
    /* Read all lines? Then return. */
    if ( (pconfig = strchr (pconfig, '\n') ) == NULL )
      return (dll);
    
    pconfig++;
  }
  
  return (dll);
}


/*
 * Returns TRUE if the item contains the category with the given name, otherwise FALSE.
 */
static short int file_category_same (DLL * fileitem, const char * category)
{
  char *line = (char *)fileitem->data;
  char *categoryname = strdup (line);
  
  /* Remove trailing/leading spaces from the current line. */
  wipe_spaces (line, 2);

  /* If the line begins with a "[", we expect a category. */
  /* The "[" MUST NOT be followed by a "]", "\n", or "\0". */
  if ( *line == '['
    && *(line + 1) != ']'
    && *(line + 1) != '\n'
    && *(line + 1) != '\0' ) {
    /* Make a copy of the categoryname. */
    strcpy (categoryname, line + 1);
      
    /* Remove the close brace from the name. */
    if ( strchr (categoryname, ']') )
      *strchr (categoryname, ']') = '\0';
    
    /* Remove trailing/leading spaces from the category name. */
    wipe_spaces (categoryname, 2);
    
    /* So is this the wanted name? Then return. */
    if ( strcmp (categoryname, category) == 0 )
      return (TRUE);
  }
  
  return (FALSE);
}


/*
 * Returns TRUE if the DLL item contains the category with the given name, otherwise FALSE.
 */
static short int line_is_category (DLL * fileitem)
{
  char *line = (char *)fileitem->data;
  
  /* Remove leading spaces from the current line. */
  wipe_spaces (line, 0);
  
  return (*line == '[');
}


/*
 * Returns TRUE if the item contains the category with the given name, otherwise FALSE.
 */
static short int file_option_same (DLL * fileitem, const char * option)
{
  char *line = (char *)fileitem->data;
  char *optionname = strdup (line);
  
  /* Remove trailing/leading spaces from the current line. */
  wipe_spaces (line, 2);
  
  /* Is this an option line? */
  if( *line != '['
    && *line != '#'
    && *line != '\n'
    && *line != '\0'
    && strchr (line, '=') != NULL ) {
    /* Make a copy of the optionname. */
    strcpy (optionname, line);
      
    /* Remove everything after the optionname. */
    *strchr (optionname, '=') = '\0';
    
    /* Remove trailing/leading spaces from the category name. */
    wipe_spaces (optionname, 2);
    
    /* So is this the wanted name? Then return TRUE. */
    if ( strcmp (optionname, option) == 0 )
      return TRUE;
  }
  
  return FALSE;
}


/*
 * Returns TRUE if the DLL contains the category with the given name, otherwise FALSE.
 */
static DLL * file_category_find (DLL * file, const char * category)
{
  DLL *fileitem = NULL;
  
  /* Parse through the DLL, looking for the correct category */
  fileitem = file;
  while ( fileitem )
  {
    /* Match? */
    if ( file_category_same (fileitem, category) )
      return (fileitem);
    
    fileitem = fileitem->next;
  }
  
  return (NULL);
}


/*
 * Remove quotation marks from a string.
 */
static void stripquotes (char * str)
{
  char * newstr = NULL;
  
  /* We need at least two quotes. */
  if ( *str != '"' || strchr (str + 1, '"') == NULL )
    return;
  
  newstr = strdup (str);
  
  /* Strip the first char. */
  strcpy (newstr, str + 1);
  strcpy (str, newstr);
  
  /* Strip the last '"' */
  *strrchr (str, '"') = '\0';
  
  free (newstr);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


/*
 * Get the requested option out of the requested category
 * out of the config variable. store its value in "value".
 * returns TRUE if successful, FALSE if not.
 */
short int configfile_option_get (const char *config, const char *category, const char *option, char *value)
{
  DLL * file = NULL;
  DLL * fileitem = NULL;
  short int rightcategory = FALSE;
  
  /* By default, the requested option has no value. */
  *value = '\0';
  
  /* Get all lines of the config into one doubly linked list. */
  if ( !(file = get_config_to_dll (config)) )
    return (FALSE);
  
  /* Find the wanted category. */
  fileitem = file;
  while ( fileitem ) {
    /* Wrong category? */
    if ( line_is_category (fileitem)
      && !file_category_same (fileitem, category) ) {
      rightcategory = FALSE;
      goto next;
    }
    
    /* Right category? */
    if ( line_is_category (fileitem)
      && file_category_same (fileitem, category) ) {
      rightcategory = TRUE;
    }
    
    /* Don't waste time. */
    if ( !rightcategory )
      goto next;
    
    /* Found the option? */
    if ( file_option_same (fileitem, option) ) {
      strcpy (value, strchr ((char *)fileitem->data, '=') + 1);
      
      wipe_spaces (value, 2);
      stripquotes (value);
      
      return (TRUE);
    }
    
next:
    fileitem = fileitem->next;
  }
  
  /* If we end up here, we haven't found anything. */
  return (FALSE);
}


/*
 * Get the requested option out of the requested category
 * out of the config variable. store its value in "value".
 * returns TRUE if successful, FALSE if not.
 */
short int configfile_option_get_integer (const char * config, const char * category, const char * option, int * value)
{
  char str_value[2048];
  
  if ( !configfile_option_get (config, category, option, str_value) )
    return (FALSE);
  *value = atoi (str_value);
  
  return (TRUE);
}


/*
 * Add an option to the config variable. If it is present, it is overwritten.
 */
short int configfile_option_add (char *config, const char *category, const char *option, const char *value)
{
  DLL * file = NULL;
  DLL * fileitem = NULL;
  DLL * categoryitem = NULL;
  DLL * optionitem = NULL;
  char * line = NULL;
  char * newline = NULL;
  char * newcategory = NULL;
  
  /* These variables must be provided by the user. */
  if ( !category || !option || !value )
    return FALSE;
  
#ifdef _DEBUG_
  printf ("--------- config ---------\n");
  printf ("%s\n", config);
  printf ("------ End of config -----\n");
#endif
  
  /* Get all lines of the config into one doubly linked list. */
  file = get_config_to_dll (config);
  
#ifdef _DEBUG_
  printf ("Configfile is %s.\n", file ? "not empty" : "empty");
#endif
  /* Find the category, if it is already in the file. */
  if ( (categoryitem = file_category_find (file, category)) ) {
    fileitem = categoryitem->next;
  }
  
  /* Parse through the current category to find the option. */
  while ( fileitem && !line_is_category (fileitem) ) {
    /* Found the option? */
    if ( file_option_same (fileitem, option) ) {
      optionitem = fileitem;
      break;
    }
    fileitem = fileitem->next;
  }
  
  /* Malloc space for the new option + 6 (The format is:'option = "value"') */
  newline = malloc (strlen (option) + strlen (value) + 6);
  sprintf (newline, "%s = \"%s\"", option, value);
  
#ifdef _DEBUG_
  printf ("Newline: %s\n", newline);
#endif
  
  /* No we do know, where the option and category are in the DLL. */
  /* If we have the option, we replace it. */
  if ( optionitem ) {
#ifdef _DEBUG_
    printf ("Have optionitem.\n");
#endif
    free (optionitem->data);
    optionitem->data = newline;
    goto cleanups;
  }
  
  /* If we have only the category, we add the new option straight below it. */
  if ( categoryitem ) {
#ifdef _DEBUG_
    printf ("Have categoryitem.\n");
#endif
    dll_insert (file, newline, dll_position (file, categoryitem));
    goto cleanups;
  }
  
#ifdef _DEBUG_
  printf ("Have neighter optionitem nor categoryitem.\n");
#endif
  
  /* Malloc space for the new category + 3 (The format is:'[category]') */
  newcategory = malloc (strlen (category) + 3);
  sprintf (newcategory, "[%s]", category);
  
#ifdef _DEBUG_
  printf ("Newcategory: %s\n", newcategory);
#endif
  
  /* If we have neighter category nor option, we add both to the end of the file. */
  /* Insert a spaces line and the option */
    file = dll_append (file, newcategory);
    file = dll_append (file, newline);
    file = dll_append (file, strdup ("\n"));
  
#ifdef _DEBUG_
  printf ("Inserted category to DLL: %s\n", newcategory);
#endif

cleanups:
  /* copy the DLL back to "config" and empty out the alloced variables. */
  *config = '\0';
  fileitem = file;
  while ( fileitem ) {
    line = (char *)fileitem->data;
    sprintf (config, "%s%s\n", config, line);
#ifdef _DEBUG_
  printf ("Inserted line to config: %s\n", line);
#endif
    free (line);
    fileitem = fileitem->next;
  }
  dll_free (file);
  
  return (TRUE);
}


short int configfile_option_add_integer (char *config, const char *category, const char *option, int * value)
{
  char str_value[50];
  int err = 0;
  
  snprintf (str_value, 49, "%i", *value);
  err = configfile_option_add (config, category, option, str_value);
  
  return (err);
}


/*
 * Removes a whole category from the config variable.
 */
short int configfile_category_remove(char *config, const char *category)
{
  char * line = NULL;
  DLL * file = NULL;
  DLL * fileitem = NULL;
  DLL * categoryitem = NULL;
  
  /* Get all lines of the config into one doubly linked list. */
  if ( !(file = get_config_to_dll (config)) )
    return TRUE;
	
  /* Find the category. */
  if ( !(categoryitem = file_category_find (file, category)) )
    return TRUE;
  
  /* Parse through the current category and remove it and every option. */
  fileitem = categoryitem->next;
  while ( fileitem && !line_is_category (fileitem) ) {
    free (fileitem->data);
    fileitem->data = NULL;
  }
  
  free (categoryitem->data);
  categoryitem->data = NULL;
  
  /* copy the DLL back to "config" and empty out the alloced variables. */
  *config = '\0';
  fileitem = file;
  while (fileitem) {
    line = (char *)fileitem->data;
    sprintf (config, "%s%s\n", config, line);
    free (line);
    fileitem = fileitem->next;
  }
  dll_free (file);
  
  return TRUE;
}


/*
 * Add a list of options to the config variable.
 */
short int configfile_set_pairs(char *config, const OVPair ** optvals)
{
  const OVPair *curpair = NULL;
  int intval = 0;
  int i = -1;

  curpair = optvals[0];
  do {
    switch (curpair->type) {
    case INTEGERU16:
      /* Convert Uint16 to int. */
      intval = *(int *) curpair->dest & 0x0000FFFF;
      configfile_option_add_integer(config, curpair->category, curpair->option, &intval);
      break;

    case INTEGER:
      configfile_option_add_integer(config, curpair->category, curpair->option, (int *) curpair->dest);
      break;

    case STRING:
      configfile_option_add(config, curpair->category, curpair->option, (char *) curpair->dest);
      break;

    default:
      fprintf(stderr, "Error: configfile_set: Invalid type option '%s/%s': Type %i.\n", curpair->category, curpair->option, curpair->type);
      break;
    }

    curpair = optvals[++i];
  } while (curpair->option);

  return (0);
}


/*
 * Get a list of options from the config variable.
 */
short int configfile_get_pairs(const char *config, const OVPair ** optvals)
{
  const OVPair *curpair = NULL;
  int i = -1, err = 0;

  curpair = optvals[0];
  do {
    switch (curpair->type) {
    case INTEGER:
    case INTEGERU16:
      err = configfile_option_get_integer(config, curpair->category, curpair->option, (int *) curpair->dest);
      break;

    case STRING:
      err = configfile_option_get(config, curpair->category, curpair->option, (char *) curpair->dest);
      break;

    default:
      fprintf(stderr, "Error: configfile_get_pairs: Invalid type option '%s/%s': Type %i.\n", curpair->category, curpair->option, curpair->type);
      break;
    }

    if (!err) {
      fprintf(stderr, "Error: configfile_get_pairs: Can't read option '%s/%s' (Type %i) from configfile.\n", curpair->category, curpair->option, curpair->type);
      return (-1);
    }

    curpair = optvals[++i];
  } while (curpair->option);

  return (0);
}


/*
 * Reads the configfile into memory
 * returns 0 if success, or >0 if an error.
 */
int configfile_read(const char *fullfilename, char *config, int maxsize)
{
  FILE * configfile = NULL;
  
  /* Open configfile. */
  if ( !(configfile = fopen (fullfilename, "rb")) )
    return (-1);
  
  /* Read file into memory. */
  memset (config, 0, maxsize);
  if ( fread (config, 1, maxsize - 1, configfile) < 3 ) {
    fclose (configfile);
    return (-2);
  }
  
  fclose (configfile);
  return (0);
}


/*
 * Writes the configfile from memory to disk.
 * returns 0 if success, or >0 if an error.
 */
int configfile_write(const char *fullfilename, char *config)
{
  FILE * configfile = NULL;
  
  /* Open configfile. */
  if ( !(configfile = fopen (fullfilename, "wb")) )
    return (-1);
  
  /* Write file. */
  if ( !fwrite (config, 1, strlen(config), configfile) ) {
    fclose (configfile);
    return (-2);
  }
  
  fclose (configfile);
  return (0);
}


/*
 * Get a list of columns to an array.
 */
void configfile_list_get (char *config, const char *category, char *array[])
{
  char option[2048];
  char value[2048];
  int curcol = 1;
  value[0] = '\0';
  
  snprintf (option, 2047, "column1");
  while ( configfile_option_get (config, category, option, value) ) {
    if ( value )
      array[curcol - 1] = strdup (value);
    array[curcol] = NULL;
    
    snprintf (option, 2047, "column%i", ++curcol);
  }
}	


/*
 * Store a list of columns from an array.
 */
void configfile_list_add (char *config, const char *category, const char *array[])
{
  char option[2048];
  int curcol = -1;
  
  while ( array[++curcol] != NULL ) {
    snprintf(option, 2047, "column%i", curcol + 1);
    configfile_option_add (config, category, option, array[curcol]);
  }
}
