/* The LinTalk Project
 * (c)2002 by Samuel Abels (sam@manicsadness.com)
 * This project's homepage is: http://software.manicsadness.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */  
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "globals.h"
#include "network.h"
#include "polling.h"
#include "users.h"

extern pthread_mutex_t mut;
extern Cfg cfg;


/***************************************************************************************
 * BELOW FOLLOW THE STATICS
 ***************************************************************************************/
/*
 * Receive a list of all known hosts from "host" and add it to our own known host list.
 */
static void add_knownhosts(GList ** users, char *phost)
{
  int sd = 0;
  char buffer_name[2048];
  char buffer_phone[2048];
  char buffer_host[2048];
  char buffer_status[2048];
  char *host = strdup(phost);

  /* Connect. */
  if ((sd = connect_host(host, LINTALK_PORT)) <= 0)
    return;

  /* Send GET ALL request. */
  if (!send_request(sd, "GET ALL\n"))
    goto error;

  /* Receive user list. */
  while (get_answer(sd, "NAME=", buffer_name, sizeof(buffer_name))) {
    /* Receive phone number, host name. */
    get_answer(sd, "PHONE=", buffer_phone, sizeof(buffer_phone));
    get_answer(sd, "HOST=", buffer_host, sizeof(buffer_host));
    get_answer(sd, "STATUS=", buffer_status, sizeof(buffer_status));

    /* Lock, because touch the user list. */
    pthread_mutex_lock(&mut);

    /* If that user is empty, or if it's me, don't add. */
    if ((buffer_name[0] == '\0' || buffer_host[0] == '\0')
      || strcmp(cfg.local.host, buffer_host) == 0) {
      pthread_mutex_unlock(&mut);
      return;
    }
    
    /* If the host is not already in our list, add it. */
    if (!user_find_from_host(cfg.users, buffer_host)) {
      User * user = user_create(buffer_name, buffer_phone, buffer_host, 0);
      user->status = user_status_char_to_int(buffer_status);
      *users = g_list_append(*users, user);
    }
    
    /* Unlock. */
    pthread_mutex_unlock(&mut);
  }

error:
  /* Close the socket and return. */
  close(sd);
  free(host);

  return;
}


/*
 * Connect every known host to grab their knownhosts.
 */
static void receive_all_known_hosts(void)
{
  GList *hosts = NULL;
  User *user = NULL;
  GList *item = NULL;
  char *host = NULL;

#ifdef _DEBUG_
  fprintf(stderr, "Creating a list of known hosts. (receive_all_known_hosts)\n");
#endif
  
  /* Lock mutex and create a list of all known hosts (must copy to be threadsafe). */
  pthread_mutex_lock(&mut);
  
  item = cfg.users;
  while (item != NULL) {
    user = item->data;
    hosts = g_list_append(hosts, strdup(user->host));
    item = item->next;
  }
  pthread_mutex_unlock(&mut);

  /* Contact every known host and receive its knownhosts. */
  item = hosts;
  while (item != NULL) {
    host = item->data;

    /* Don't connect to myself. */
    if (strcmp(host, cfg.local.host) == 0) {
      item = item->next;
      continue;
    }
#ifdef _DEBUG_
    fprintf(stderr, "Grabbing hostnames from host: %s. (receive_all_known_hosts)\n", host);
    
#endif
    /* Add knownhosts to our list. */
    add_knownhosts(&cfg.users, host);

#ifdef _DEBUG_
    fprintf(stderr, "Grabbing hostnames from host: %s DONE! (receive_all_known_hosts)\n", host);
#endif
    
    free(host);

    item = item->next;
  }

  g_list_free(hosts);

#ifdef _DEBUG_EXCESSIVE_
  fprintf(stderr, "End of function (receive_all_known_hosts)\n");
#endif
}
/***************************************************************************************
 * END OF STATICS
 ***************************************************************************************/


void *pollingloop(void *psd)
{
  GList *item = NULL;
  User *user = NULL;
  int sl = 0;

  /* Update our user list. */
  receive_all_known_hosts();

#ifdef _DEBUG_EXCESSIVE_
  fprintf(stderr, "Currently known hosts (pollingloop):\n");
  pthread_mutex_lock(&mut);
  item = cfg.users;
  while (item != NULL) {
    user = item->data;

    fprintf(stderr, "%s\n", user->name);
    fprintf(stderr, "%s\n", user->phone);
    fprintf(stderr, "%s\n", user->host);
    item = item->next;
  }
  pthread_mutex_unlock(&mut);
#endif

  /* Now we have all known hosts in our userlist. */
  while (!cfg.shutdown) {
    /* Poll all known hosts, and receive the known host list of all new hosts. */
    pthread_mutex_lock(&mut);

    item = cfg.users;
    while (item) {
      user = item->data;

      /* Don't connect to myself. */
      if (strcmp(user->host, cfg.local.host) == 0) {
        item = item->next;
        continue;
      }

      /* Is the host newly added? Then receive its complete host list. */
      if (user->new) {
        user->new = FALSE;
        
        pthread_mutex_unlock(&mut);
        add_knownhosts(&cfg.users, user->host);
        pthread_mutex_lock(&mut);
      }

      /* Update user details. */
      pthread_mutex_unlock(&mut);
      receive_who(user, user->host);
      pthread_mutex_lock(&mut);

      item = item->next;
    }

    pthread_mutex_unlock(&mut);

#ifdef _DEBUG_EXCESSIVE_
    fprintf(stderr, "Currently known hosts (pollingloop):\n");
    pthread_mutex_lock(&mut);
    item = cfg.users;
    while (item != NULL) {
      user = item->data;
      fprintf(stderr, "%s\n", user->name);
      fprintf(stderr, "%s\n", user->phone);
      fprintf(stderr, "%s\n", user->host);
      item = item->next;
    }
    pthread_mutex_unlock(&mut);
    fprintf(stderr, "GUI updating\n");
#endif
    
    /* Wait a while before polling each host the next time. */
    /* While waiting, we must still check if the server is running. */
#ifdef _DEBUG_
    printf("Poll interval hit.\n");
#endif
    sl = 0;
    while (sl < POLLSLEEP && !cfg.shutdown) {
      sleep(1);
      sl++;
    }
  }
  pthread_exit(0);
}
